import os
from django.core import management
from fabric.api import run, sudo, local, put, env, settings, abort, cd
from fabric.contrib.console import confirm

INSTALL_DIR = "/home/ubuntu/"
INSTALL_PATH = os.path.join(INSTALL_DIR, 'helipad')


def pip_install_non_debian_items():
    """pip install packages not in Ubuntu distribution."""

    sudo("pip install django-jsonfield")


def apt_get_everything():
    """apt-get everything required to be a helipad."""

    sudo("apt-get install -y bzr apache2 libapache2-mod-wsgi \
          python-django postgresql python-psycopg2 python-pip")


def restart_postgres():
    """Restart our fav DB server."""

    sudo("service postgresql restart")


def create_postgres_user():
    """Create 'ubuntu' postgres user."""

    # does the user exist already?
    result = sudo("psql postgres -tAc \"SELECT 1 FROM pg_roles WHERE rolname='ubuntu'\"",
                  user='postgres')
    if result != "1":
        with settings(warn_only=True):
            sudo("createuser -U postgres -s ubuntu -P", user='postgres')
            # we sometimes warn on this but it's ok IMO
            sudo("createdb ubuntu", user='postgres')


def database_exists():
    """Does the database exist?"""

    result = sudo("psql -tAc \"SELECT * FROM pg_database WHERE datname='helipad';\"",
                  user='postgres')
    return result != ""


def destroy_database():
    """Destroy postgresql helipad database."""

    with settings(warn_only=True):
        # we could warn on this if it doesn't exist (but we don't expect to)
        sudo("dropdb helipad", user='postgres')


def create_database():
    """Create a fresh helipad db for user ubuntu."""

    sudo("createdb -E utf8 -O ubuntu helipad -T template0",
         user='postgres')


def adjust_pg_hba_conf():
    """We need md5 password checking for localhost."""
 
    # TODO: this sed dinna work under fabric WTF:
    # sudo sed 's/^\(local\s*all\s*all\s*\)\(.*\)$/\1md5/' /etc/postgresql/9.1/main/pg_hba.conf
    # . . . so we're just copying a baked file
    put("./pg_hba.conf", "/etc/postgresql/9.1/main/", use_sudo=True)
    restart_postgres()


def setup_database():
    """Create postgresql helipad user and database."""

    adjust_pg_hba_conf()
    if database_exists():
        destroy_database()
    create_postgres_user()
    create_database()
    restart_postgres()


def checkout_helipad(branch="lp:helipad", location=INSTALL_DIR):
    """Checkout branch into 'helipad' in the user's home dir.

    Bazaar and fabric don't get along so we can't just bzr branch,
    instead cp from a local checkout.
    """

    install_path = os.path.join(location, 'helipad')
    # TODO: this could just be in /tmp
    tmp_install_path = os.path.join(os.path.abspath(os.path.curdir),
                                    'tmp', 'helipad')
    local("rm -rf {}".format(tmp_install_path))
    local("mkdir {}".format(os.path.dirname(tmp_install_path)))
    sudo("rm -rf {}".format(install_path))
    local("bzr branch {} {}".format(branch, tmp_install_path))
    put(tmp_install_path, os.path.dirname(install_path), use_sudo=True)
    local("rm -rf {}".format(tmp_install_path))
    local("rmdir {}".format(os.path.dirname(tmp_install_path)))
    # set perms so that www-data can access
    sudo("chown -R ubuntu:helipad {}".format(install_path))
    sudo("chmod -R g+rw {}".format(install_path))


def restart_apache():
    """Restart our fav webserver."""

    sudo("service apache2 restart")


def configure_apache():
    """Configure apache2: copy our site config, set up perms, etc."""

    apache_config_path = os.path.join(INSTALL_DIR, 'helipad', 'deployment', 'helipad')
    sudo("cp {} /etc/apache2/sites-available/".format(apache_config_path))
    sudo("adduser www-data helipad")
    sudo("a2ensite helipad")
    restart_apache()


def setup_users():
    """Create a helipad group and add ubuntu to it.

    This facilitates apache access.
    """

    with settings(warn_only=True):
        sudo("addgroup helipad")
        sudo("adduser ubuntu helipad")


def copy_local_settings(install_path=INSTALL_PATH):
    """Copy production settings into place."""

    sudo("cp {} {}".format(os.path.join(install_path,
                                       'deployment',
                                       'settings_production.py'),
                           os.path.join(install_path,
                                        "helipad",
                                        "local_settings.py")),
         user='ubuntu')


def collect_static_media():
    """Use the Django utility to collect all of our static files."""

    with cd(INSTALL_PATH):
        sudo("python manage.py collectstatic --noinput", user='ubuntu')


def deploy(branch="lp:helipad"):
    """Push this branch to the production site."""

    checkout_helipad(branch)
    configure_apache()
    copy_local_settings()
    collect_static_media()
    with cd(INSTALL_PATH):
        run("python manage.py syncdb")
        run("python manage.py migrate")


# TODO:
def deploy_staging():
    pass


def setup(branch="lp:helipad"):
    """Setup the helipad site on a new server.

    Should be needed rarely, drops db.
    """
    
    apt_get_everything()
    pip_install_non_debian_items()
    setup_users()
    setup_database()
    deploy()
