
import copy
import os
import sys
import logging
import tempfile
import shutil
import fnmatch
from subprocess import call
from django.core.management.base import BaseCommand, CommandError
from optparse import make_option
from psqa.util.command_utils import (slurp_cu2d_config,
                                     slurp_project_from_stack)

from c2dconfigutils.cu2dStackLookup import StackLookup
from c2dconfigutils.c2dconfigutils import (load_default_cfg, load_stack_cfg)


logger = logging.getLogger("helipad")

# TODO please confirm/correct this and/or put it in settings.py
CU2D_REPO = "lp:cupstream2distro-config"
CU2D_CONFIG_FILENAME_EXTENSION = "*.cfg"
CU2D_IGNORE_KEYS = ['ci_default', 'daily_release_default', 'projects',
                    'to_transition']


# NOTE that we only use the last two terms of the path!
CONFIG_BLACKLIST = ['phablet/ubuntu-touch-coreapps.cfg']


def config_is_on_blacklist(config_path):
    """Return a bool if we find this filepath in the blacklist."""

    # get the last two terms of the offered path, e.g. 'phablet/apps.cfg'
    config_path = os.path.sep.join(config_path.split(os.path.sep)[-2:])
    return config_path in CONFIG_BLACKLIST


def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for filename in files:
            if fnmatch.fnmatch(filename, pattern):
                filename = os.path.join(root, filename)
                yield filename


class Command(BaseCommand):
    help = "Slurp all cu2d configs in a repo and deposit to database."

    def _create_stack(self, stackcfg, release):
        """Creates a temporary dict for Cu2dStack creation"""
        stack_name = '{}.{}'.format(release, stackcfg['name'])
        # DEBUGGING: Used to limit the stacks being slurped:
        #if stack_name != 'phablet.apps':
        #    return None
        stack = dict()
        stack['name'] = stack_name
        stack['release'] = release
        stack['series'] = stackcfg.get('series', '')
        stack['ppa'] = stackcfg.get('ppa', '')
        logger.debug("Slurping stack: {}".format(stack_name))
        slurp_cu2d_config(stack)
        return stack_name

    def _create_project(self, lookup, stackcfg, release, stack_name,
                        project_name):
        """Creates a temporary dict for PsqaProject creation"""
        project = dict()
        project['release'] = release
        project['stack'] = stack_name
        project['name'] = "{}.{}".format(stack_name, project_name)
        ci_jobs = lookup.get_ci_name(stackcfg, project_name)
        if len(ci_jobs) > 0:
            project['ci'] = ci_jobs[0]
        else:
            project['ci'] = None
        al_jobs = lookup.get_autolanding_name(stackcfg, project_name)
        if len(al_jobs) > 0:
            project['autolanding'] = al_jobs[0]
        else:
            project['autolanding'] = None
        logger.debug("Slurping project: {}".format(project_name))
        slurp_project_from_stack(project)

    def handle(self, *args, **options):
        repo_dir = tempfile.mkdtemp()
        orig_dir = os.getcwd()
        os.chdir(repo_dir)
        call(["bzr", "co", CU2D_REPO])
        os.chdir(orig_dir)
        config_path = os.path.join(repo_dir, 'cupstream2distro-config', 'ci')
        master_default_config = load_default_cfg(config_path)
        stack_base = os.path.join(repo_dir, 'cupstream2distro-config',
                                  'stacks')
        for config_path in find_files(stack_base,
                                      CU2D_CONFIG_FILENAME_EXTENSION):
            if config_is_on_blacklist(config_path):
                logger.debug("Skipping cu2d config at {} due to blacklist match.".format(config_path))
                continue
            default_config = copy.deepcopy(master_default_config)
            logger.debug("Slurping cu2d config at {}".format(config_path))
            stackcfg = load_stack_cfg(config_path, default_config)
            
            # Extract release name from the file path:
            # /tmp/tmp_dir/stack/{release}/stack.cfg
            (head, tail) = os.path.split(config_path)
            (head, release) = os.path.split(head)
            stack_name = self._create_stack(stackcfg, release)
            if stack_name is None:
                continue
            
            lookup = StackLookup()
            try:
                projects = lookup.get_projects(stackcfg)
            except TypeError:
                logger.debug("No projects found in {}".format(config_path))
                continue
            logger.debug("Projects: %s" % (projects))
            for project_name in projects:
                self._create_project(lookup, stackcfg, release, stack_name,
                                     project_name)
        shutil.rmtree(repo_dir)
