import logging
import yaml
from django.db import models, IntegrityError
import util.extractor as extractor
from launchpadlib.launchpad import Launchpad
from jsonfield import JSONField


logger = logging.getLogger("helipad")

# TODO? these could move to settings?
AUTOLANDING_JOB_REGEX_TEMPLATE = "{}-autolanding"
CI_JOB_REGEX_TEMPLATE = "{}-ci"
DAILY_JOB_REGEX_TEMPLATE = "{}-daily"

BLACKLIST_JOBS = [
    'generic-dput',
    'generic-land',
    'generic-update_mp',
    'phablet-dput',
    'phablet-land',
]

class JenkinsJobManager(models.Manager):

    def create_from_url(self, url):
        # TODO: uniqueness constraint in db, catch exception
        # Blacklist infrastructure jobs
        for name in BLACKLIST_JOBS:
            if name in url:
                return None
        extractor_job = extractor.JenkinsJob(url)
        try:
            jenkins_job = self.create(
                name=extractor_job.name,
                url=url,
                buildable=extractor_job.buildable,
                in_queue=extractor_job.in_queue,
                last_build=extractor_job.last_build,
                last_completed_build=extractor_job.last_completed_build,
                last_failed_build=extractor_job.last_failed_build,
                last_successful_build=extractor_job.last_successful_build,
                parameters=extractor_job.parameters)
        except IntegrityError:
            # uniqueness constraint violated, instead update the object
            logger.debug("updating job")
            jenkins_job = JenkinsJob.objects.get(name=extractor_job.name)
            jenkins_job.in_queue = extractor_job.in_queue
            jenkins_job.last_build = extractor_job.last_build
            jenkins_job.last_completed_build = extractor_job.last_completed_build
            jenkins_job.last_failed_build = extractor_job.last_failed_build
            jenkins_job.last_successful_build = extractor_job.last_successful_build
            jenkins_job.parameters = extractor_job.parameters
            jenkins_job.save()
        for build in extractor_job.builds:
            jenkins_build = JenkinsBuild.objects.create_from_url(build.url, jenkins_job)
            # DEBUGGING: Hack, only return 1 build to speed things up
            #break
        for downstream_extractor_job in extractor_job.downstream_jobs:
            # TODO: protect against circular refs, just sayin'
            downstream_jenkins_job = self.create_from_url(
                downstream_extractor_job.url)
            if downstream_jenkins_job:
                downstream_jenkins_job.upstream_job = jenkins_job
                downstream_jenkins_job.save()
        return jenkins_job


    def create_from_name(self, name):
        logger.debug("creating JenkinsJob from name {}".format(name))
        return self.create_from_url(extractor.get_jenkins_job_url_from_job_name(name))


    def get_by_name_regex(self, regex):
        """Return get on a regex search on the name field."""

        return JenkinsJob.objects.get(name__regex=regex)


    def get_autolanding_job(self, name):
        """Find an autolanding job using project name and regex."""

        return self.get_by_name_regex(
            AUTOLANDING_JOB_REGEX_TEMPLATE.format(name))


    def get_ci_job(self, name):
        """Find a CI job using project name and regex."""

        return self.get_by_name_regex(
            CI_JOB_REGEX_TEMPLATE.format(name))


    def get_daily_job(self, name):
        """Find a daily job using project name and regex."""

        return self.get_by_name_regex(
            DAILY_JOB_REGEX_TEMPLATE.format(name))


class JenkinsJob(models.Model):
    """A Jenkins job."""

    #enabled = models.BooleanField()
    name = models.CharField(max_length=256, unique=True)
    url = models.URLField()
    project = models.ForeignKey('PsqaProject', related_name='jenkins_jobs', null=True)
    upstream_job = models.ForeignKey('JenkinsJob', related_name='downstream_jobs', null=True)
    buildable = models.BooleanField()
    in_queue = models.BooleanField()
    last_build = models.IntegerField(null=True)
    last_completed_build = models.IntegerField(null=True)
    last_failed_build = models.IntegerField(null=True)
    last_successful_build = models.IntegerField(null=True)
    parameters = JSONField(max_length=256, null=True)

    def line_coverage(self):
        return 20.0

    def branch_coverage(self):
        return 20.0

    objects = JenkinsJobManager()

    def __unicode__(self):
        return self.name


class PsqaProjectGroup(models.Model):
    """A tag for project grouping."""


class PsqaProjectManager(models.Manager):
    def create_from_name(self, name):
        # TODO really we should overload create() itself
        # link to our related objects
        psqa_project = self.create(name=name)
        try:
            launchpad_project = LaunchpadProject.objects.get(name=name)
        except LaunchpadProject.DoesNotExist:
            launchpad_project = LaunchpadProject.objects.create_from_name(name=name)
        psqa_project.launchpad_projects.add(launchpad_project)
        psqa_project.save()
        return psqa_project

    def create_from_config(self, config):
        psqa_project, created = self.get_or_create(name=config['name'])
        psqa_project.stack = Cu2dStack.objects.get(name=config['stack'])
        psqa_project.release = Release.objects.get(name=config['release'])
        job = JenkinsJob.objects.create_from_name(config['ci'])
        if job is not None:
            job.project = psqa_project
            job.save()
        psqa_project.ci_primary_job = job
        job = JenkinsJob.objects.create_from_name(config['autolanding'])
        if job is not None:
            job.project = psqa_project
            job.save()
        psqa_project.autolanding_primary_job = job
        psqa_project.save()

class PsqaProject(models.Model):
    """Our first-class object which links Jenkins to Launchpad."""

    name = models.CharField(max_length=256, unique=True)

    # Job relationships
    release = models.ForeignKey('Release', null=True)
    stack = models.ForeignKey('Cu2dStack', null=True)
    ci_primary_job = models.ForeignKey(
        'JenkinsJob', related_name='ci_primary', null=True)
    autolanding_primary_job = models.ForeignKey(
        'JenkinsJob', related_name='autolanding_primary', null=True)

    staging_ppa = models.CharField(max_length=256, null=True)
    active = models.NullBooleanField(null=True)
    responsible = models.CharField(max_length=256, null=True, blank=True)
    comment = models.TextField(max_length=1024, null=True, blank=True)

    # more like "tag"; also consider recursive dependencies instead
    super_project = models.NullBooleanField(null=True)

    # these may belong to the Launchpad project?
    maintainer = models.CharField(max_length=256, null=True, blank=True)
    driver = models.CharField(max_length=256, null=True, blank=True)
    bug_supervisor = models.CharField(max_length=256, null=True, blank=True)
    security_contact = models.CharField(max_length=256, null=True, blank=True)
    project_group = models.CharField(max_length=256, null=True, blank=True)

    # we may or may not own a Launchpad project
    launchpad_projects = models.ManyToManyField('LaunchpadProject', null=True)

    def line_coverage(self):
        last_build = self.last_build()
        if last_build is not None:
            if last_build.coverage_report is not None:
                return last_build.coverage_report.line_rate
            else:
                return None
        else:
            return None

    def branch_coverage(self):
        last_build = self.last_build()
        if last_build is not None:
            if last_build.coverage_report is not None:
                return last_build.coverage_report.branch_rate
            else:
                return None
        else:
            return None

    def last_build(self):
        try:
            jenkins_job = self.autolanding_job
        except:
            # FIXME: what is this exception?
            return None
        try:
            last_build = JenkinsBuild.objects.filter(jenkins_job=jenkins_job).order_by('timestamp')[0]
        except Exception, e:
            return None
        return last_build

    def _get_autolanding_job(self):
        return JenkinsJob.objects.get_autolanding_job(self.name)
    autolanding_job = property(_get_autolanding_job)

    def _get_ci_job(self):
        return JenkinsJob.objects.get_ci_job(name)
    ci_job = property(_get_ci_job)

    def _get_daily_job(self):
        return JenkinsJob.objects.get_daily_job(name)
    daily_job = property(_get_daily_job)

    # TODO
    #def _get_coverage_report(self):
    #    build = self._get_last_autolanded_build()
    #
    #coverage_report = property(_get_coverage_report)

    # TODO
    #test_report = property()

    objects = PsqaProjectManager()

    def __unicode__(self):
        return self.name


class LaunchpadProjectManager(models.Manager):

    def create_from_name(self, name):
        lp = Launchpad.login_with('helipad', 'production')
        lp_project = lp.projects[name]
        # TODO check for PsqaProjects which may be related
        return self.create(name=name,
                           url=lp_project.web_link)


class LaunchpadProject(models.Model):
    """Corresponds to a top-level Launchpad project."""

    name = models.CharField(max_length=256)
    # TODO need that lp:foo link in here somewhere
    url = models.URLField()
    
    objects = LaunchpadProjectManager()

    def __unicode__(self):
        return self.name


class Cu2dStackManager(models.Manager):

    def create_from_config(self, config):
        series, created = Series.objects.get_or_create(name=config['series'])
        release, created = Release.objects.get_or_create(name=config['release'])
        stack, created = Cu2dStack.objects.get_or_create(name=config['name'])
        stack.series = series
        stack.ppa = config['ppa']
        stack.release = release
        stack.save()
        return stack


class Cu2dStack(models.Model):
    """Canonical upstream to distro stack.

    An agglomeration of projects often released together.
    """
    
    name = models.CharField(max_length=64)
    series = models.ForeignKey('Series', null=True)
    ppa = models.CharField(max_length=64, null=True)
    release = models.ForeignKey('Release', null=True)

    objects = Cu2dStackManager()

    def __unicode__(self):
        return self.name
    

class JenkinsTestReport(models.Model):
    run = models.ForeignKey('JenkinsRun')
    fail_count = models.IntegerField()
    pass_count = models.IntegerField()
    skip_count = models.IntegerField()
    duration = models.IntegerField() # milliseconds
    # TODO can we get the report itself?


class JenkinsCoverageReportManager(models.Manager):

    def create_from_cobertura_xml(self, xml_text, jenkins_build):
        # TODO actual coverage rates please
        coverage_report = self.create(
            jenkins_build=jenkins_build,
            xml_text=xml_text,
            line_rate=0.0,#extractor_run.coverage_report.line_rate,
            branch_rate=0.0)#extractor_run.coverage_report.branch_rate)
        return coverage_report


class JenkinsCoverageReport(models.Model):
    jenkins_build = models.ForeignKey('JenkinsBuild', null=True)
    xml_text = models.TextField()
    line_rate = models.FloatField()
    branch_rate = models.FloatField()
    # TODO moar data

    objects = JenkinsCoverageReportManager()


class JenkinsFlavor(models.Model):
    name = models.CharField(max_length=64)

    def __unicode__(self):
        return self.name


class Release(models.Model):
    name = models.CharField(max_length=64)

    def __unicode__(self):
        return self.name


class Series(models.Model):
    name = models.CharField(max_length=64)

    def __unicode__(self):
        return self.name


class JenkinsRunParameter(models.Model):
    key = models.ForeignKey('JenkinsBuildParameter')
    value = models.CharField(max_length=256)

    def __unicode__(self):
        return "{}: {}".format(key, value)


class JenkinsBuildParameter(models.Model):
    name = models.CharField(max_length=64)

    def __unicode__(self):
        return self.name


class JenkinsResult(models.Model):
    name = models.CharField(max_length=64)

    def __unicode__(self):
        return self.name


class JenkinsRunManager(models.Manager):

    def create_from_url(self, url, jenkins_build):
        extractor_run = extractor.JenkinsRun(url)
        flavor, created = JenkinsFlavor.objects.get_or_create(name=extractor_run.flavor)
        series, created = Series.objects.get_or_create(name=extractor_run.series)
        try:
            result, created = JenkinsResult.objects.get_or_create(name=extractor_run.result)
        except IntegrityError, e:
            # no builds yet, IntegrityError on None name
            pass
        try:
            logger.debug("creating JenkinsRun at {}".format(url))
            jenkins_run = self.create(
                jenkins_build=jenkins_build,
                number=extractor_run.number,
                name=extractor_run.name,
                url=url,
                duration=extractor_run.duration.seconds,
                timestamp=extractor_run.timestamp,
                flavor=flavor,
                series=series,
                result=result,
                parameters=extractor_run.parameters)
        except IntegrityError:
            logger.debug("updating JenkinsRun at {}".format(url))
            # maybe the run was in progress before, e.g.
            jenkins_run = JenkinsRun.objects.get(url=url)
            jenkins_run.jenkins_build = jenkins_build
            jenkins_run.number = extractor_run.number
            jenkins_run.number=extractor_run.number
            jenkins_run.name=extractor_run.name
            jenkins_run.url=url
            jenkins_run.duration=extractor_run.duration.seconds
            jenkins_run.timestamp=extractor_run.timestamp
            jenkins_run.flavor=flavor
            jenkins_run.series=series
            jenkins_run.result=result
            jenkins_run.parameters=extractor_run.parameters
            jenkins_run.save()
        logger.debug("importing coverage reports associated with JenkinsRun at {}".format(url))
        try:
            jenkins_run.coverage_report = JenkinsCoverageReport.objects.create(
                run = jenkins_run,
                xml_text=extractor_run.coverage_report.xml_text,
                line_rate=extractor_run.coverage_report.line_rate,
                branch_rate=extractor_run.coverage_report.branch_rate)
            jenkins_run.save()
        except EnvironmentError:
            # no data was found at the coverage report url
            pass
        logger.debug("importing test reports associated with JenkinsRun at {}".format(url))
        try:
            jenkins_run.test_report = JenkinsTestReport.objects.create(
                run = jenkins_run,
                fail_count=extractor_run.test_report.fail_count,
                pass_count=extractor_run.test_report.pass_count,
                skip_count=extractor_run.test_report.skip_count,
                duration=extractor_run.test_report.duration.seconds)
            jenkins_run.save()
        except EnvironmentError:
            # no data was found at the test report url
            pass
        except extractor.ExtractorException, e:
            logger.error(e)

        return jenkins_run


class JenkinsRun(models.Model):

    jenkins_build = models.ForeignKey('JenkinsBuild',
                                      related_name='runs',
                                      null=True)
    number = models.IntegerField(null=True)
    name = models.CharField(max_length=256)
    url = models.URLField(unique=True)
    duration = models.IntegerField() # seconds
    flavor = models.ForeignKey('JenkinsFlavor')
    series = models.ForeignKey('Series')
    result = models.ForeignKey('JenkinsResult')
    parameters = JSONField(max_length=256, null=True)
    coverage_report = models.ForeignKey('JenkinsCoverageReport', null=True)
    test_report = models.ForeignKey('JenkinsTestReport', null=True)
    timestamp = models.DateTimeField()

    objects = JenkinsRunManager()

    def __unicode__(self):
        return self.name


class JenkinsBuildManager(models.Manager):

    def create_from_url(self, url, jenkins_job):
        """Job is a Django entity to which to relate."""
        for name in BLACKLIST_JOBS:
            if name in url:
                return None
        extractor_build = extractor.JenkinsBuild(url)
        if extractor_build.result is None:
            # This build is actively running, no need to slurp it
            return None
        try:
            result, created = JenkinsResult.objects.get_or_create(name=extractor_build.result)
        except IntegrityError, e:
            # no builds yet, IntegrityError on None name
            pass
        try:
            logger.debug("creating JenkinsBuild at {}".format(url))
            jenkins_build = self.create(
                jenkins_job=jenkins_job,
                url=url,
                number=extractor_build.number,
                timestamp=extractor_build.timestamp,
                result=result)
        except IntegrityError:
            logger.debug("updating JenkinsBuild at {}".format(url))
            # maybe the run was in progress before, e.g.
            jenkins_build = JenkinsBuild.objects.get(url=url)
            jenkins_build.number = extractor_build.number
            jenkins_build.timestamp = extractor_build.timestamp
            jenkins_build.result = result
        logger.debug("importing runs associated with JenkinsBuild at {}".format(url))
        for run in extractor_build.runs:
            jenkins_run = JenkinsRun.objects.create_from_url(run.url,
                                                             jenkins_build)
        # get downstream builds
        downstream_builds = extractor.get_downstream_builds_dict_from_build_url(url)
        # make sure we have all the jobs represented before linking against builds
        logger.debug("downstream_builds = {}".format(downstream_builds))
        for job_name in [build.get('name') for build in downstream_builds]:
            logger.debug("do we have JenkinsJob {} yet?".format(job_name))
            try:
                jenkins_job = JenkinsJob.objects.get(name=job_name)
            except JenkinsJob.DoesNotExist:
                jenkins_job = JenkinsJob.objects.create_from_name(name=job_name)
        # now get related builds and link 'em
        for downstream_build in downstream_builds:
            name = downstream_build['name']
            number = downstream_build['number']
            logger.debug("testing for {} #{}".format(name, number))
            try:
                jenkins_job = JenkinsJob.objects.get(name=name)
            except JenkinsJob.DoesNotExist:
                continue
            try:
                downstream_jenkins_build = JenkinsBuild.objects.get(
                    jenkins_job=jenkins_job,
                    number=number)
            except JenkinsBuild.DoesNotExist:
                downstream_jenkins_build = JenkinsBuild.objects.create_from_name_and_number(name, number)
            if downstream_jenkins_build:
                jenkins_build.downstream_builds.add(downstream_jenkins_build)
        jenkins_build.save()
        # get a Cobertura coverage report, if it exists
        try:
            cobertura_report = extractor.get_cobertura_report_from_build_url(url)
            jenkins_coverage_report = JenkinsCoverageReport.objects.create_from_cobertura_xml(cobertura_report, jenkins_build)
            jenkins_build.coverage_report = jenkins_coverage_report
            jenkins_build.save()
        except extractor.ExtractorException, e:
            # no biggie, we just don't have a Cobertura report
            logger.info("No Cobertura data: {}".format(e))

        return jenkins_build


    def create_from_name_and_number(self, name, number):
        for blacklist in BLACKLIST_JOBS:
            if blacklist in name:
                return None
        logger.debug("creating JenkinsBuild for {} #{}".format(name, number))
        jenkins_job = JenkinsJob.objects.get(name=name)
        return self.create_from_url(
            extractor.get_jenkins_build_url_from_job_name_and_build_number(
                name, number), jenkins_job)


class JenkinsBuild(models.Model):
    """A Jenkins build."""

    # TODO I think these are properties (?)
    #coverage_report
    #flavors
    #parameters
    #series
    #test_reports
    jenkins_job = models.ForeignKey('JenkinsJob', related_name='builds', null=True)
    number = models.IntegerField()
    timestamp = models.DateTimeField()
    url = models.URLField(unique=True)
    result = models.ForeignKey('JenkinsResult')
    downstream_builds = models.ManyToManyField('JenkinsBuild')
    coverage_report = models.ForeignKey(JenkinsCoverageReport, null=True)

    objects = JenkinsBuildManager()

    def __unicode__(self):
        return self.url
