
import re
from mock import patch, MagicMock
from urlparse import urljoin
from django.test import TestCase
from psqa.util.extractor import (get_build_console_text,
                                 get_downstream_builds_dict_from_build_url,
                                 get_downstream_builds_dict_from_build_console_text,
                                 compose_dict_from_downstream_build_name)
from helipad.settings import JENKINS_URL


CONSOLE_TEXT = """Started by user ps-jenkins
Building on master
Waiting for the completion of mir-clang-raring-amd64-build
mir-clang-raring-amd64-build #84 completed. Result was SUCCESS
Waiting for the completion of mir-android-raring-i386-build
mir-android-raring-i386-build #201 completed. Result was SUCCESS
Waiting for the completion of mir-quantal-amd64-autolanding
mir-quantal-amd64-autolanding #100 completed. Result was SUCCESS
Build step 'Trigger/call builds on other projects' changed build result to SUCCESS
Waiting for the completion of generic-land
generic-land #6943 completed. Result was SUCCESS
Build was marked for publishing on https://jenkins.qa.ubuntu.com/
Triggering a new build of mir-vm-autolanding-build #306
Finished: SUCCESS
"""

DOWNSTREAM_BUILDS = [
    {'name': 'mir-clang-raring-amd64-build',
     'number': 84,
     'url': urljoin(JENKINS_URL, 'job/mir-clang-raring-amd64-build/84/')},
    {'name': 'mir-android-raring-i386-build',
     'number': 201,
     'url': urljoin(JENKINS_URL, 'job/mir-android-raring-i386-build/201/')},
    {'name': 'mir-quantal-amd64-autolanding',
     'number': 100,
     'url': urljoin(JENKINS_URL, 'job/mir-quantal-amd64-autolanding/100/')},
    {'name': 'generic-land',
     'number': 6943,
     'url': urljoin(JENKINS_URL, 'job/generic-land/6943/')},
]


class TestGetConsoleTextForBuild(TestCase):
    """Integration test."""

    def test_vanilla(self):
        build_url = urljoin(JENKINS_URL, "job/mir-autolanding/99/")
        result = get_build_console_text(build_url)
        self.assertEqual(CONSOLE_TEXT, result)


class TestGetDownstreamBuildsFromBuildConsoleText(TestCase):
    def test_vanilla(self):
        result = get_downstream_builds_dict_from_build_console_text(CONSOLE_TEXT)
        self.assertEqual(DOWNSTREAM_BUILDS, result)


class TestGetDownstreamBuildsFromBuildUrl(TestCase):

    def test_vanilla(self):
        get_build_console_text_patch = MagicMock(return_value=CONSOLE_TEXT)
        get_downstream_builds_dict_from_build_console_text_patch = MagicMock(
            return_value=DOWNSTREAM_BUILDS)
        with patch('psqa.util.extractor.get_build_console_text',
                   get_build_console_text_patch), \
                  patch('psqa.util.extractor.get_downstream_builds_dict_from_build_console_text',
                        get_downstream_builds_dict_from_build_console_text_patch):
            result = get_downstream_builds_dict_from_build_url(
                urljoin(JENKINS_URL, 'job/mir-autolanding/'))
        self.assertEqual(DOWNSTREAM_BUILDS, result)


class TestMatchDownstreamBuildLine(TestCase):
    
    line = "mir-clang-raring-amd64-build #84 completed. Result was SUCCESS"
    build_regex = re.compile("^(?P<name>.*) #(?P<number>\d+) completed\. Result was .*$")
    
    def test_we_find_something(self):
        self.assertIsNotNone(self.build_regex.match(self.line))

    def test_match_is_equal_to_build_name(self):
        result = self.build_regex.match(self.line)
        if result is not None:
            self.assertEqual('mir-clang-raring-amd64-build',
                             result.group('name'))
        else:
            self.fail("no regular expression match")

    def test_match_is_equal_to_build_number(self):
        result = self.build_regex.match(self.line)
        if result is not None:
            self.assertEqual('84',
                             result.group('number'))
        else:
            self.fail("no regular expression match")


class TestMatchDownstreamBuildText(TestCase):

    build_regex = re.compile("^(?P<name>.*) #(?P<number>\d+) completed\. Result was .*$",
                             re.MULTILINE)

    def test_we_find_something(self):
        result = self.build_regex.search(CONSOLE_TEXT)
        self.assertIsNotNone(result)

    def test_correct_number_of_result_groups(self):
        result = self.build_regex.findall(CONSOLE_TEXT)
        if result is not None:
            self.assertEqual(4, len(result))
        else:
            self.fail("no result groups found")


class TestComposeDictFromDownstreamBuildName(TestCase):

    def test_vanilla(self):
        result = compose_dict_from_downstream_build_name(
            'mir-clang-raring-amd64-build', 84)
        expected = {'name': 'mir-clang-raring-amd64-build',
                    'number': 84,
                    'url': urljoin(JENKINS_URL,
                                   'job/mir-clang-raring-amd64-build/84/')}
        self.assertEqual(result, expected)
