import os
from mock import MagicMock, patch
import datetime
import json
import unittest
from urlparse import urljoin
from psqa.util.extractor import JenkinsBuild
from helipad.settings import JENKINS_URL


DATA_PATH = "psqa/tests/data"
BUILD_FILE = "nux-mbs-autolanding.jenkins.build"


class JenkinsBuildTestCase(unittest.TestCase):
    """Unit tests for class JenkinsBuild"""

    def setUp(self):
        url = urljoin(JENKINS_URL, "/view/mbs/job/nux-mbs-autolanding/98/")
        data = self._get_data("nux-mbs-autolanding-98.json")
        self.build = JenkinsBuild(url, data)
        self.amd64_data = self._get_data("nux-mbs-autolanding-98-amd64.json")
        self.armhf_data = self._get_data("nux-mbs-autolanding-98-armhf.json")
        self.i386_data = self._get_data("nux-mbs-autolanding-98-i386.json")
        self.report_data = self._get_data("nux-mbs-autolanding-98-amd64-testReport.json")
        self.amd64_cov = \
            self._get_text("nux-mbs-autolanding-98-amd64.coverage.xml")
        self.armhf_cov = \
            self._get_text("nux-mbs-autolanding-98-armhf.coverage.xml")
        self.i386_cov = \
            self._get_text("nux-mbs-autolanding-98-i386.coverage.xml")

    def _get_data(self, filename):
        return json.loads(self._get_text(filename))

    def _get_text(self, filename):
        with open(os.path.join(DATA_PATH, filename)) as in_file:
            return in_file.read()

    def test_name(self):
        self.assertEqual(self.build.name, "nux-mbs-autolanding #98")

    def test_number(self):
        self.assertEqual(self.build.number, 98)

    def test_run_count(self):
        runs = None
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [self.armhf_data, self.amd64_data,
                                     self.i386_data]
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            runs = self.build.runs
        self.assertEqual(len(runs), 3)

    def test_data_none(self):
        url = urljoin(JENKINS_URL, "/view/mbs/job/compiz-mbs-autolanding/193/")
        data = self._get_data("compiz-mbs-autolanding-193.json")
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [data]
        value = None
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            build = JenkinsBuild(url)
            value = "%s" % (build)
        self.assertEqual(len(value), 3214)

    def test_duration(self):
        self.assertEqual(self.build.duration.seconds,
                         datetime.timedelta(seconds=11, minutes=77).seconds)

    def test_flavors(self):
        flavors = None
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [self.armhf_data, self.amd64_data,
                                     self.i386_data]
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            flavors = self.build.flavors
        self.assertEqual(flavors, set(['armhf', 'amd64', 'i386']))

    def test_parameters(self):
        self.assertEqual(len(self.build.parameters), 19)
        self.assertEqual(self.build.parameters['project'], 'nux')
        self.assertEqual(self.build.parameters['target_branch'], 'lp:nux')

    def test_series(self):
        series = None
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [self.armhf_data, self.amd64_data,
                                     self.i386_data]
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            series = self.build.series
        self.assertEqual(series, set(['raring']))

    def test_result(self):
        self.assertEqual(self.build.result, "SUCCESS")

    def test_result_null(self):
        url = urljoin(JENKINS_URL, "/view/mbs/job/nux-mbs-autolanding/98/")
        data = self._get_data("compiz-mbs-autolanding-193.json")
        build = JenkinsBuild(url, data)
        self.assertEqual(build.result, None)

    def test_building(self):
        self.assertEqual(self.build.building, False)

    def test_building_true(self):
        url = urljoin(JENKINS_URL, "/view/mbs/job/nux-mbs-autolanding/98/")
        data = self._get_data("compiz-mbs-autolanding-193.json")
        build = JenkinsBuild(url, data)
        self.assertEqual(build.building, True)

    def test_test_report(self):
        test_report = None
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [self.armhf_data, self.amd64_data,
                                     self.i386_data, self.report_data,
                                     self.report_data, self.report_data]
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            test_report = self.build.test_report
        self.assertEqual(test_report.pass_count, 393)

    def test_coverage_report(self):
        coverage_report = None
        get_data_mock = MagicMock()
        get_data_mock.side_effect = [self.armhf_data, self.amd64_data,
                                     self.i386_data]
        get_text_mock = MagicMock()
        get_text_mock.side_effect = [self.armhf_cov, self.amd64_cov,
                                     self.i386_cov]
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            with patch('psqa.util.extractor.get_text_from_url', get_text_mock):
                coverage_report = self.build.coverage
                self.assertEqual(coverage_report.lines, 3854)
                self.assertAlmostEqual(coverage_report.line_rate, float(0.28489880))
                self.assertAlmostEqual(coverage_report.branch_rate, float(0.58794946))
