"""Tests for module JenkinsCoverage"""

from mock import MagicMock, patch
import os.path
import unittest
from psqa.util.extractor import (JenkinsCoverage,
                                 get_coverage_from_url,
                                 get_coverage_from_daily,
                                 get_coverage_for_arch,
                                 get_coverage_from_job)

DATA_PATH = "psqa/tests/data"
COVERAGE_FILE = "coverage.xml"


class JenkinsCoverageTestCase(unittest.TestCase):
    """Unit tests for class JenkinsCoverage"""

    def setUp(self):
        """Setup for all test cases"""
        url = 'http://10.97.0.6:8080/job/nux-mbs-autolanding/build=pbuilder,distribution=raring,flavor=amd64/98/artifacts/work/results/coverage.xml'
        xml_text = None
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            xml_text = xml_file.read()
        self.jc = JenkinsCoverage(url, xml_text)

    def tearDown(self):
        """Tear down for all test cases"""
        pass

    def test_str(self):
        """Verify xml tree is converted to a string"""
        self.assertIn("<packages>", "%s" % (self.jc))
        self.assertIn("</packages>", "%s" % (self.jc))

    def test_get_attribute(self):
        self.assertAlmostEqual(float(self.jc.get_attribute('line-rate')),
                               float(0.28452967))
        self.assertAlmostEqual(float(self.jc.get_attribute('branch-rate')),
                               float(0.58794946))

    def test_get_coverage_from_xml(self):
        coverage = self.jc.get_coverage_from_xml()
        self.assertAlmostEqual(float(coverage['line-rate']),
                               float(0.28452967))
        self.assertAlmostEqual(float(coverage['branch-rate']),
                               float(0.58794946))

    def test_get_packages(self):
        packages = self.jc.get_packages()
        self.assertEqual(packages, ['NuxCore..libs.FileManager',
                                    'NuxCore..libs',
                                    'NuxCore..libs.Character',
                                    'NuxCore..libs.Math'])

    def test_lines(self):
        self.assertEqual(self.jc.lines, 3859)
        self.assertEqual(self.jc.hits, 1098)
        self.assertAlmostEqual(float(self.jc.hits) /
                               float(self.jc.lines),
                               float(self.jc.get_attribute('line-rate')))

    def test_branches(self):
        self.assertEqual(self.jc.branches, 324)
        self.assertEqual(self.jc.taken, 605)
        self.assertEqual(self.jc.total, 1029)
        self.assertAlmostEqual(float(self.jc.taken) /
                               float(self.jc.total),
                               float(self.jc.get_attribute('branch-rate')))

    def test_get_counts_lines_for_package(self):
        coverage = self.jc.get_counts(package='NuxCore..libs')
        self.assertEqual(coverage.lines, 2200)
        self.assertEqual(coverage.hits, 1071)
        self.assertAlmostEqual(float(coverage.hits) /
                               float(coverage.lines),
                               float(0.48681818))

class GenericFake():
    def __init__(self):
        pass


class JenkinsCoverageFromURLTestCase(unittest.TestCase):
    """Unit tests for get_coverage_from_url"""

    def test_get_coverage_for_arch(self):
        job = None
        fake_request = GenericFake()
        fake_request.status_code = 200
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job_url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
            job = get_coverage_for_arch(job_url,
                                        'raring',
                                        arches)
        self.assertAlmostEqual(float(job.get_coverage_from_xml()['line-rate']),
                               float(0.28452967))

    def test_get_coverage_for_arch_fail(self):
        job = None
        fake_request = GenericFake()
        fake_request.status_code = 0
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job_url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
            with self.assertRaises(EnvironmentError) as err:
                job = get_coverage_for_arch(job_url,
                                            'raring',
                                            arches)
            self.assertIn("No coverage file", "%s" % (err.exception))

    def test_get_coverage_from_daily(self):
        job = None
        fake_request = GenericFake()
        fake_request.status_code = 200
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job_url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
            job = get_coverage_from_daily(job_url)
        self.assertAlmostEqual(float(job.get_coverage_from_xml()['line-rate']),
                               float(0.28452967))

    def test_get_coverage_from_url(self):
        job = None
        fake_request = GenericFake()
        fake_request.status_code = 200
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job_url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
            job = get_coverage_from_url(job_url,
                                        series='raring',
                                        arch_order=arches)
        self.assertAlmostEqual(float(job.get_coverage_from_xml()['line-rate']),
                               float(0.28452967))

    def test_get_coverage_from_url_fail(self):
        job = None
        fake_request = GenericFake()
        fake_request.status_code = 0
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job_url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
            with self.assertRaises(EnvironmentError) as err:
                job = get_coverage_from_url(job_url)
            self.assertIn("No coverage file", "%s" % (err.exception))

class JenkinsCoverageFromJenkinsJobTestCase(unittest.TestCase):
    """Unit tests for get_coverage_from_url"""

    def test_get_coverage_from_jenkins_job(self):
        job = None
        fake_request = GenericFake()
        fake_server = GenericFake()
        fake_job = GenericFake()
        url = "http://10.97.0.6:8080/view/mbs/job/nux-mbs-autolanding"
        fake_job.baseurl = url
        fake_request.status_code = 200
        with open(os.path.join(DATA_PATH, COVERAGE_FILE)) as xml_file:
            fake_request.text = xml_file.read()
        mock_get = MagicMock(return_value=fake_request)
        mock_get_job = MagicMock(return_value=fake_job)
        fake_server.get_job = mock_get_job
        with patch('requests.get', mock_get):
            arches = ['amd64', 'i386', 'armhf']
            job = get_coverage_from_job(fake_server,
                                        'nux-mbs-autolanding',
                                        series='raring',
                                        arch_order=arches)
        self.assertAlmostEqual(float(job.get_coverage_from_xml()['line-rate']),
                               float(0.28452967))
