from mock import MagicMock, patch
import datetime
import json
import os.path
import unittest
from psqa.util.extractor import JenkinsRun

"""Tests for module JenkinsRun"""

DATA_PATH = "psqa/tests/data"
DATA_DICT = {'amd64': {'url': 'http://10.97.0.6:8080/job/nux-mbs-autolanding/build=pbuilder,distribution=raring,flavor=amd64/98',
                       'file': 'nux-mbs-autolanding-98-amd64.json',
                      },
             'armhf': {'url': 'http://10.97.0.6:8080/job/nux-mbs-autolanding/build=panda-pbuilder,distribution=raring,flavor=armhf/98',
                       'file': 'nux-mbs-autolanding-98-armhf.json',
                      },
             'i386': {'url': 'http://10.97.0.6:8080/job/nux-mbs-autolanding/build=pbuilder,distribution=raring,flavor=i386/98',
                      'file': 'nux-mbs-autolanding-98-i386.json',
                     },
             'compiz': {'url':
                        'http://10.97.0.6:8080/job/compiz-mbs-autolanding/build=panda-pbuilder,distribution=raring,flavor=armhf/193',
                       'file': 'compiz-mbs-autolanding-193-armhf.json',
                      },
             'testReport': {'url': 'http://10.97.0.6:8080/job/nux-mbs-autolanding/build=pbuilder,distribution=raring,flavor=amd64/98/testReport',
                       'file': 'nux-mbs-autolanding-98-amd64-testReport.json',
                      },
            }


class JenkinsRunTestCase(unittest.TestCase):
    """Unit tests for class JenkinsRun"""

    def setUp(self):
        pass

    def _get_data(self, key):
        with open(os.path.join(DATA_PATH,
                               DATA_DICT[key]['file'])) as in_file:
            return {'json': json.loads(in_file.read()),
                    'url': DATA_DICT[key]['url']}

    def test_init(self):
        name = 'amd64'
        data = self._get_data(name)
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.url, DATA_DICT[name]['url'])
        self.assertEqual(jr.data[u'fullDisplayName'],
                         u'nux-mbs-autolanding \xbb pbuilder,raring,amd64 #98')

    def test_str(self):
        data = self._get_data('compiz')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertIn("<JenkinsRun: ", "%s" % jr)
        self.assertIn("compiz-mbs-autolanding", "%s" % jr)

    def test_data_none(self):
        data = self._get_data('armhf')
        get_data_mock = MagicMock(return_value=data['json'])
        value = None
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            jr = JenkinsRun(data['url'])
            value = "%s" % (jr)
        self.assertEqual(len(value), 2691)

    def test_duration(self):
        data = self._get_data('i386')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.duration.seconds,
                         datetime.timedelta(seconds=39,minutes=23).seconds)

    def test_flavor(self):
        flavor = 'amd64'
        data = self._get_data(flavor)
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.flavor, flavor)

    def test_number(self):
        data = self._get_data('i386')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.number, 98)

    def test_series(self):
        data = self._get_data('armhf')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.series, 'raring')

    def test_result(self):
        data = self._get_data('amd64')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.result, "SUCCESS")

    def test_result_null(self):
        data = self._get_data("compiz")
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.result, None)

    def test_building(self):
        data = self._get_data('i386')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.building, False)

    def test_building_true(self):
        data = self._get_data("compiz")
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(jr.building, True)

    def test_report(self):
        data = self._get_data('amd64')
        test_data = self._get_data('testReport')
        get_data_mock = MagicMock(return_value=test_data['json'])
        with patch('psqa.util.extractor.get_json_from_url', get_data_mock):
            jr = JenkinsRun(data['url'], data['json'])
            self.assertEqual(jr.test_report.pass_count, 131)

    def test_parameters(self):
        data = self._get_data('amd64')
        jr = JenkinsRun(data['url'], data['json'])
        self.assertEqual(len(jr.parameters), 19)
        self.assertEqual(jr.parameters['project'], 'nux')
