import datetime
import logging
import re
import json
import requests
from lxml import etree
from urlparse import urljoin
from django.utils.timezone import utc
from helipad.settings import JENKINS_URL


"""Classes and utilities to extract QA data from jenkins"""

logger = logging.getLogger('helipad')


class ExtractorException(Exception):
    pass


class JenkinsObject(object):
    """The base class for other jenkins objects with common attributes"""

    def __init__(self, url, data=None):
        self.url = url
        self._data = data

    @property
    def data(self):
        if self._data is None:
            self._data = get_json_from_url(self.url)
        return self._data

    @property
    def name(self):
        return self.data['displayName']


class JenkinsJob(JenkinsObject):

    """Provides data access for a jenkins job"""

    def __init__(self, url, data=None):
        """Creates the job object"""
        super(JenkinsJob, self).__init__(url, data)
        self._builds = None
        self._downstream_jobs = None
        self._parameters = None

    def __str__(self):
        return "<JenkinsJob: %s>" % (self.data)

    @property
    def builds(self):
        if self._builds is None:
            self._builds = list()
            try:
                for b in self.data['builds']:
                    build = JenkinsBuild(b['url'])
                    self._builds.append(build)
            except TypeError, e:
                # no builds
                return []
        return self._builds

    @property
    def downstream_jobs(self):
        if self._downstream_jobs is None:
            self._downstream_jobs = list()
            for j in self._data['downstreamProjects']:
                job = JenkinsJob(j['url'])
                self._downstream_jobs.append(job)
        return self._downstream_jobs

    @property
    def buildable(self):
        return self._data['buildable']

    @property
    def in_queue(self):
        return self._data['inQueue']

    @property
    def last_build(self):
        try:
            return int(self._data['lastBuild']['number'])
        except TypeError:
            # no build
            return None

    @property
    def last_completed_build(self):
        try:
            return int(self._data['lastCompletedBuild']['number'])
        except TypeError:
            # no build
            return None

    @property
    def last_failed_build(self):
        try:
            return int(self._data['lastFailedBuild']['number'])
        except TypeError:
            # no build
            return None

    @property
    def last_successful_build(self):
        try:
            return int(self._data['lastSuccessfulBuild']['number'])
        except TypeError:
            # no build
            return None

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = dict()
            parameters = self.data['actions'][0]['parameterDefinitions']
            for p in parameters:
                name = p['name']
                self._parameters[name] = p['defaultParameterValue']['value']
        return self._parameters


class JenkinsBuild(JenkinsObject):

    """Provides data access for a jenkins build"""

    def __init__(self, url, data=None):
        """Creates the build object"""
        super(JenkinsBuild, self).__init__(url, data)
        self._coverage_report = None
        self._duration = None
        self._flavors = None
        self._parameters = None
        self._runs = None
        self._series = None
        self._test_report = None
        self._timestamp = None

    def __str__(self):
        return "<JenkinsBuild: %s>" % (self.data)

    @property
    def coverage(self):
        """Aggregates the individual test run coverage reports"""
        if self._coverage_report is None:
            report = JenkinsCoverage()
            for run in self.runs:
                report.add(run.coverage_report)
            report.lines = report.lines / len(self.runs)
            report.hits = report.hits / len(self.runs)
            report.branches = report.branches / len(self.runs)
            report.taken = report.taken / len(self.runs)
            report.total = report.total / len(self.runs)
            self._coverage_report = report
        return self._coverage_report

    @property
    def duration(self):
        if self._duration is None:
            self._duration = datetime.timedelta(
                milliseconds=int(self.data['duration']))
        return self._duration

    @property
    def timestamp(self):
        if self._timestamp is None:
            timestamp = int(self.data['timestamp']) / 1000
            self._timestamp = datetime.datetime.fromtimestamp(timestamp).replace(tzinfo=utc)
        return self._timestamp

    @property
    def flavors(self):
        """Returns a set containing all flavors found in the build"""
        if self._flavors is None:
            self._flavors = set()
            for run in self.runs:
                self._flavors.add(run.flavor)
        return self._flavors

    @property
    def number(self):
        """Returns the number of this build"""
        return int(self.data['number'])

    @property
    def runs(self):
        """Returns a dictionary of flavors with run statistics"""
        if self._runs is None:
            try:
                runs = self.data['runs']
            except KeyError:
                logger.info("No runs for build {}.".format(self.name))
                return []
            self._runs = list()
            for r in runs:
                # The api may return a run for ever configuration in the
                # build matrix, throw out those that don't match the build
                # number.
                if r['number'] == self.data['number']:
                    url = r['url']
                    data = get_json_from_url(url)
                    self._runs.append(JenkinsRun(url, data))
        return self._runs

    @property
    def series(self):
        """Returns a set containing all flavors found in the build"""
        if self._series is None:
            self._series = set()
            for run in self.runs:
                self._series.add(run.series)
        return self._series

    @property
    def test_report(self):
        """Aggregates the individual test run results for the build"""
        if self._test_report is None:
            self._test_report = JenkinsTestReport()
            for run in self.runs:
                self._test_report.add(run.test_report)
        return self._test_report

    @property
    def data(self):
        if self._data is None:
            self._data = get_json_from_url(self.url)
        return self._data

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = dict()
            parameters = self.data['actions'][0]['parameters']
            for p in parameters:
                name = p['name']
                self._parameters[name] = p['value']
        return self._parameters

    @property
    def name(self):
        return self.data['fullDisplayName']

    @property
    def building(self):
        return self.data['building']

    @property
    def result(self):
        return self.data['result']


class JenkinsRun(JenkinsObject):

    FLAVORS = ['armel', 'armhf', 'amd64', 'i386']
    SERIES = ['precise', 'quantal', 'raring']

    def __init__(self, url, data=None):
        """Creates the run object"""
        super(JenkinsRun, self).__init__(url, data)
        self._duration = None
        self._flavor = None
        self._number = None
        self._parameters = None
        self._series = None
        self._coverage_report = None
        self._test_report = None
        self._timestamp = None

    def __str__(self):
        return "<JenkinsRun: %s>" % (self.data)

    @property
    def data(self):
        if self._data is None:
            self._data = get_json_from_url(self.url)
        return self._data

    @property
    def name(self):
        return self.data['fullDisplayName']

    @property
    def duration(self):
        if self._duration is None:
            self._duration = datetime.timedelta(
                milliseconds=int(self.data['duration']))
        return self._duration

    @property
    def timestamp(self):
        if self._timestamp is None:
            timestamp = int(self.data['timestamp']) / 1000
            self._timestamp = datetime.datetime.fromtimestamp(timestamp).replace(tzinfo=utc)
        return self._timestamp

    @property
    def flavor(self):
        if self._flavor is None:
            for flavor in self.FLAVORS:
                if flavor in self.url:
                    self._flavor = flavor
        return self._flavor

    @property
    def number(self):
        if self._number is None:
            self._number = int(self.data['number'])
        return self._number

    @property
    def series(self):
        if self._series is None:
            for series in self.SERIES:
                if series in self.url:
                    self._series = series
        return self._series

    @property
    def building(self):
        return self.data['building']

    @property
    def result(self):
        return self.data['result']

    @property
    def test_report(self):
        try:
            if self._test_report is None:
                url = "/".join([self.url, self.data['actions'][3]['urlName']])
                data = get_json_from_url(url)
                self._test_report = JenkinsTestReport(url, data)
            return self._test_report
        except KeyError, e:
            raise ExtractorException("Failed to get test report: {}".format(e))

    @property
    def coverage_report(self):
        if self._coverage_report is None:
            url = "/".join([self.url, "artifact/work/results/coverage.xml"])
            text = get_text_from_url(url)
            self._coverage_report = JenkinsCoverage(url, text)
        return self._coverage_report

    @property
    def parameters(self):
        if self._parameters is None:
            self._parameters = dict()
            parameters = self.data['actions'][0]['parameters']
            for p in parameters:
                name = p['name']
                self._parameters[name] = p['value']
        return self._parameters


class JenkinsTestReport(JenkinsObject):

    """Provides methods and attributes from a jenkins test report."""

    def __init__(self, url=None, data=None):
        """Creates a test report object"""
        super(JenkinsTestReport, self).__init__(url, data)
        self.fail_count = 0
        self.pass_count = 0
        self.skip_count = 0
        self.duration = datetime.timedelta(0)
        if self._data is not None:
            self.fail_count = int(self._data['failCount'])
            self.pass_count = int(self._data['passCount'])
            self.skip_count = int(self._data['skipCount'])
            self.duration = datetime.timedelta(
                microseconds=int(float(self._data['duration'] * 1000)))
        self.children = set()

    def add(self, report):
        if report not in self.children:
            self.children.add(report)
            self.fail_count += report.fail_count
            self.pass_count += report.pass_count
            self.skip_count += report.skip_count
            if self.duration < report.duration:
                self.duration = report.duration

coverage_url_part = {'arch':
                     '/build=pbuilder,distribution=%s,flavor=%s'
                     '/lastSuccessfulBuild/artifact/work/results/coverage.xml',
                     'noarch':
                     '/lastSuccessfulBuild/artifact/work/results/coverage.xml',
                    }


class JenkinsCoverage(JenkinsObject):

    """Provides methods and attributes for a jenkins code coverage report."""

    def __init__(self, url=None, data=None):
        """Creates the coverage object"""
        super(JenkinsCoverage, self).__init__(url, data)
        self.xml_text = data
        self.root = None
        self.branches = 0
        self.hits = 0
        self.lines = 0
        self.taken = 0
        self.total = 0
        if self.xml_text is not None:
            self.root = etree.XML(self.xml_text)
            self.get_counts()
        self.children = set()

    def __str__(self):
        """Provides a string representation of the xml tree"""
        return "<JenkinsCoverage %s>" % (self.xml_text)

    @property
    def line_rate(self):
        return float(self.hits) / float(self.lines)

    @property
    def branch_rate(self):
        return float(self.taken) / float(self.total)

    def add(self, coverage):
        if coverage not in self.children:
            self.children.add(coverage)
            self.branches += coverage.branches
            self.hits += coverage.hits
            self.lines += coverage.lines
            self.taken += coverage.taken
            self.total += coverage.total

    def get_attribute(self, attribute, root=None):
        if root is None:
            root = self.root
        return root.get(attribute)

    def get_coverage_from_xml(self):
        coverage = dict()
        for attr in ['line-rate', 'branch-rate']:
            coverage[attr] = self.get_attribute(attr)
        return coverage

    def get_packages(self):
        packages = list()
        for element in self.root[1]:
            packages.append(self.get_attribute('name', element))
        return packages

    def get_counts(self, package=None):
        """Records detailed coverage stats from the coverage xml"""
        collect = True
        obj = self
        for element in self.root.iter():
            if package is not None and element.tag == 'package':
                if self.get_attribute('name', element) == package:
                    collect = True
                    package_dict = dict(lines=0, hits=0,
                                        branches=0, taken=0, total=0)
                    obj = type('CobeturaCoveragePackage',
                               (JenkinsCoverage,),
                               package_dict)
                else:
                    collect = False
            if collect and element.tag == 'line':
                obj.lines += 1
                if self.get_attribute('hits', element) != '0':
                    obj.hits += 1
                if self.get_attribute('branch', element) == 'true':
                    obj.branches += 1
                    coverage = self.get_attribute('condition-coverage',
                                                  element)
                    coverage = coverage[coverage.find('(') + 1:-1]
                    (taken, total) = coverage.split('/')
                    obj.taken += int(taken)
                    obj.total += int(total)
        return obj


def get_coverage_for_arch(job_url, series, arch_order):
    url = None
    for arch in arch_order:
        url = job_url + coverage_url_part['arch'] % (series, arch)
        coverage = requests.get(url)
        if coverage.status_code == 200:
            break

    if coverage.status_code != 200:
        raise EnvironmentError('No coverage file')
    return JenkinsCoverage(url, "%s" % (coverage.text))


def get_coverage_from_daily(job_url):
    url = job_url + coverage_url_part['noarch']
    coverage = requests.get(url)

    if coverage.status_code != 200:
        raise EnvironmentError('No coverage file')
    return JenkinsCoverage(url, "%s" % (coverage.text))


def get_coverage_from_url(job_url, series=None, arch_order=None):
    if arch_order:
        coverage = get_coverage_for_arch(job_url, series, arch_order)
    else:
        coverage = get_coverage_from_daily(job_url)
    return coverage


def get_coverage_from_job(server, job_name, series=None, arch_order=None):
    """Retreives the coverage data from a jenkins server and job"""
    job = server.get_job(job_name)
    return get_coverage_from_url(job.baseurl, series, arch_order)


def get_json_from_url(url):
    url = "/".join([url, "api", "json"])
    page = requests.get(url)
    if page.status_code != 200:
        raise EnvironmentError("No json data at %s" % (url))
    return json.loads(page.text)


def get_text_from_url(url):
    page = requests.get(url)
    if page.status_code != 200:
        raise EnvironmentError("No data at %s" % (url))
    return page.text


def get_build_console_text(build_url):
    """Return a block of text from a Jenkins build."""

    build_console_text_url = build_url + "/consoleText"
    # TODO whatif this errors
    return requests.get(build_console_text_url).text


def compose_dict_from_downstream_build_name(name, number):
    """Given what we know, compose a dict of infos from the build name."""

    return {'name': name,
            'number': number,
            'url': urljoin(JENKINS_URL,
                           "job/{}/{}/".format(name, number))}


def get_downstream_builds_dict_from_build_console_text(text):
    """Return a list of dicts with build name, number, and url."""

    downstream_builds = []
    build_regex = re.compile("^(?P<name>.*) #(?P<number>\d+) completed\. Result was .*$",
                             re.MULTILINE)
    results = build_regex.finditer(text)
    for result in results:
        downstream_builds.append(compose_dict_from_downstream_build_name(
                result.group('name'), int(result.group('number'))))
    return downstream_builds


def get_downstream_builds_dict_from_build_url(url):
    """Return a dict of build names, numbers, and urls."""

    text = get_build_console_text(url)
    return get_downstream_builds_dict_from_build_console_text(text)


def get_jenkins_job_url_from_job_name(job_name):
    """Compose a URL given the Jenkins Job name."""

    return urljoin(JENKINS_URL, "job/{}/".format(job_name))


def get_jenkins_build_url_from_job_name_and_build_number(job_name, build_number):
    """Compose a URL for a Jenkins Build given its job name and number."""

    return urljoin(JENKINS_URL, "job/{}/{}/".format(job_name,
                                                    build_number))

def compose_build_cobertura_url(build_url):
    return "{}/artifact/work/results/coverage.xml".format(build_url)


def get_cobertura_report_from_build_url(build_url):
    """Return a bool based on attempt to req the Cobertura URL."""

    build_cobertura_url = compose_build_cobertura_url(build_url)
    page = requests.get(build_cobertura_url)
    if page.status_code != 200:
        raise ExtractorException("No data at {}".format(build_cobertura_url))
    return page.text
